/*
 *   News Peruser Copyright (c) 1996-1998 James Bailie
 *   ==================================================================
 *
 *   News Peruser is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2, or (at
 *   your option) any later version.
 *
 *   News Peruser is distributed in the hope that it will be useful, but
 *   WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *   General Public License for more details.
 *
 *   Although News Peruser is licensed under the Free Software
 *   Foundation's GNU General Public License, Peruser is not produced
 *   by, nor is it endorsed by the Free Software Foundation. The Free
 *   Software Foundation is not responsible for developing,
 *   distributing, or supporting Peruser in any way. Anyone may place
 *   software they own the copyright to, under the GNU General Public
 *   License.
 *
 *   The GNU General Public License is included in the News Peruser 
 *   distribution archive in a file called COPYING. If you do
 *   not have a copy of the license, you can download one from
 *   ftp://prep.ai.mit.edu, or you can write to the Free Software
 *   Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *   =====================================================================
 */

#include "libnews.h"

/*
 * Removes articles marked as seen from the specified group's spoolfile.
 */

int ln_pack_spool( char *group_name, char folder )
{
  char buffer[ LN_BUFFER_SIZE ], 
    spool_path[ LN_BUFFER_SIZE ], temp_path[ LN_BUFFER_SIZE ], 
    read_path[ LN_BUFFER_SIZE ], temp_read_path[ LN_BUFFER_SIZE ], 
    ids_path[ LN_BUFFER_SIZE ], temp_ids_path[ LN_BUFFER_SIZE ],
    *home;

  FILE *spool, *read, *temp, *temp_read, *ids, *temp_ids;


  home = getenv( "HOME" );
  snprintf( spool_path, LN_BUFFER_SIZE, "%s/.peruser_spool/%s", 
	    home, group_name );
  if (( spool = fopen( spool_path, "r" )) == NULL )
    {
      snprintf( ln_error_message, LN_BUFFER_SIZE,
		"libnews: ln_pack_spool: could not open file %s.",
		spool_path );
      return -2;
    }

  snprintf( read_path, LN_BUFFER_SIZE, "%s:read", spool_path );
  if (( read = fopen( read_path, "r" )) == NULL )
    {
      fclose( spool );
      snprintf( ln_error_message, LN_BUFFER_SIZE,
		"libnews: ln_pack_spool: could not open file %s.",
		read_path );
      return -2;
    }

  snprintf( temp_path, LN_BUFFER_SIZE, "%s/.peruser_spool/%s:temp", 
	    home, group_name );
  if (( temp = fopen( temp_path, "w" )) == NULL )
    {
      fclose( spool );
      fclose( read );
      snprintf( ln_error_message, LN_BUFFER_SIZE,
		"libnews: ln_pack_spool: could not open file %s.",
		temp_path );
      return -1;
    }

  snprintf( temp_read_path, LN_BUFFER_SIZE, "%s/.peruser_spool/%s:read:temp", 
	    home, group_name );
  if (( temp_read = fopen( temp_read_path, "w" )) == NULL )
    {
      fclose( spool );
      fclose( read );
      fclose( temp );
      snprintf( ln_error_message, LN_BUFFER_SIZE,
		"libnews: ln_pack_spool: could not open file %s.",
		temp_read_path );
      return -1;
    }

  if ( folder )
    {
      snprintf( ids_path, LN_BUFFER_SIZE, "%s/.peruser_spool/%s:ids",
		home, group_name );
      if (( ids = fopen( ids_path, "r" )) == NULL )
	{
	  fclose( spool );
	  fclose( read );
	  fclose( temp );
	  fclose( temp_read );
	  snprintf( ln_error_message, LN_BUFFER_SIZE,
		    "libnews: ln_pack_spool: could not open file %s.",
		    ids_path );
	  return -1;
	}

      strcpy( temp_ids_path, ids_path );
      strcat( temp_ids_path, ".tmp" );

      if ( ( temp_ids = fopen( temp_ids_path, "w" )) == NULL )
	{
	  fclose( spool );
	  fclose( temp );
	  fclose( read );
	  fclose( temp_read );
	  fclose( ids );
	  snprintf( ln_error_message, LN_BUFFER_SIZE, 
		    "libnews: ln_pack_spool: could not open file %s.",
		    temp_ids_path );
	  return -1;
	}
    }

  while( fgets( buffer, LN_BUFFER_SIZE, read ) != NULL )
    {
      if ( buffer[ 0 ] == 'u' )
	{
	  do
	    {
	      if ( fgets( buffer, LN_BUFFER_SIZE, spool ) == NULL )
		{
		  snprintf( ln_error_message, LN_BUFFER_SIZE,
			    "libnews: ln_pack_spool: "
			    "premature end of spool %s.",
			    group_name );
		  return -1;
		}
	      fputs( buffer, temp );
	    }
	  while( strncmp( buffer, ".\r\n", 3 ));

	  fputs( "u\n", temp_read );

	  if ( folder )
            fputs( fgets( buffer, LN_BUFFER_SIZE, ids ), temp_ids );
	}
      else
	{
	  do
            if ( fgets( buffer, LN_BUFFER_SIZE, spool ) == NULL )
	      {
		snprintf( ln_error_message, LN_BUFFER_SIZE,
			  "libnews: ln_pack_spool: premature end of spool %s.",
			  group_name );
		return -1;
	      }
	  while( strncmp( buffer, ".\r\n", 3 ));

	  if ( folder )
            fgets( buffer, LN_BUFFER_SIZE, ids );
	}
    }

  fclose( spool );
  fclose( read );
  fclose( temp );
  fclose( temp_read );
  rename( temp_path, spool_path );
  rename( temp_read_path, read_path );

  if ( folder )
    {
      fclose( ids );
      fclose( temp_ids );
      rename( temp_ids_path, ids_path );
    }

  return 0;
}
