/*
 *   News Peruser Copyright (c) 1996-1998 James Bailie
 *   ==================================================================
 *
 *   News Peruser is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2, or (at
 *   your option) any later version.
 *
 *   News Peruser is distributed in the hope that it will be useful, but
 *   WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *   General Public License for more details.
 *
 *   Although News Peruser is licensed under the Free Software
 *   Foundation's GNU General Public License, Peruser is not produced
 *   by, nor is it endorsed by the Free Software Foundation. The Free
 *   Software Foundation is not responsible for developing,
 *   distributing, or supporting Peruser in any way. Anyone may place
 *   software they own the copyright to, under the GNU General Public
 *   License.
 *
 *   The GNU General Public License is included in the News Peruser 
 *   distribution archive in a file called COPYING. If you do
 *   not have a copy of the license, you can download one from
 *   ftp://prep.ai.mit.edu, or you can write to the Free Software
 *   Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *   =====================================================================
 */

#include "libnews.h"

#include<unistd.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>

/*
 * Posts messages to server connected to stream.
 */

int ln_post_messages( FILE *stream, char *server, unsigned int *count )
{
   char *home, buffer[ LN_BUFFER_SIZE ], second_buffer[ LN_BUFFER_SIZE ];
   FILE *spool, *posted, *temp, *rejects = NULL;


   *count = 0;

   home = getenv( "HOME" );
   snprintf( buffer, LN_BUFFER_SIZE, "%s/.peruser_spool/%s-FOLLOW-UPS", 
             home, server );
   if (( spool = fopen( buffer, "r" )) == NULL )
   {
      if ( errno == ENOENT )
      {
         snprintf( ln_error_message, LN_BUFFER_SIZE,
                   "libnews: ln_post_messages: no follow-ups on file for %s",
                   server );
         return -2;
      }
         
      snprintf( ln_error_message, LN_BUFFER_SIZE,
                "libnews: ln_post_messages: could not open file %s.",
                buffer );
      return -1;
   }

   if (( buffer[ 0 ] = fgetc( spool )) == EOF )
   {
      fclose( spool );
      strcpy( ln_error_message,
              "libnews: ln_post_messages: no outgoing articles queued." );
      return -2;
   }
   ungetc( buffer[ 0 ], spool );

   snprintf( buffer, LN_BUFFER_SIZE, "%s/.peruser_spool/%s-POSTED", home,
             server );
   if (( posted = fopen( buffer, "a" )) == NULL )
   {
      fclose( spool );
      snprintf( ln_error_message, LN_BUFFER_SIZE,
                "libnews: ln_post_messages: could not open file %s.",
                buffer );
      return -1;
   }

   snprintf( buffer, LN_BUFFER_SIZE, "%s/.peruser_spool/.temp", home );
   if (( temp = fopen( buffer, "w+" )) == NULL )
   {
      fclose( spool );
      fclose( posted );
      snprintf( ln_error_message, LN_BUFFER_SIZE,
                "libnews: ln_post_messages: could not open file %s.",
                buffer );
      return -1;
   }

   for( ; ; )
   {
      if (( buffer[ 0 ] = fgetc( spool )) == EOF )
         break;
      ungetc( buffer[ 0 ], spool );

      fputs( "post\r\n", stream );
      if ( strncmp( ln_fgets( buffer, LN_BUFFER_SIZE, stream ),
                    "340", 3 ))
      {
         fclose( spool );
         fclose( posted );
         fclose( temp );
         snprintf( ln_error_message, LN_BUFFER_SIZE,
                   "libnews: ln_post_messages: %s.",
                   buffer );
         return -2;
      }

      do
      {
         if ( fgets( buffer, LN_BUFFER_SIZE, spool ) == NULL )
         {
            fclose( temp );
            fclose( spool );
            fclose( posted );
            snprintf( ln_error_message, LN_BUFFER_SIZE,
                      "libnews: ln_post_messages: "
                      "premature end of ~/.peruser_spool/%s-FOLLOW-UPS.",
                      server );
            return -1;
         }

         fputs( buffer, temp );
         fputs( buffer, stream );
      }
      while( strncmp( buffer, ".\r\n", 3 ));

      ln_fgets( buffer, LN_BUFFER_SIZE, stream );
      if ( strncmp( buffer, "240", 3 ))
      {
         snprintf( buffer, LN_BUFFER_SIZE, "%s/.peruser_spool/.rejects",
                   home, server );
         if (( rejects = fopen( buffer, "a" )) == NULL )
         {
            fclose( spool );
            fclose( temp );
            fclose( posted );
            snprintf( ln_error_message, LN_BUFFER_SIZE, 
                      "libnews: ln_post_messages: "
                      "could not open file: %s.", buffer );
            return -1;
         }
         
         rewind( temp );
         while( fgets( buffer, LN_BUFFER_SIZE, temp ) != NULL )
            fputs( buffer, rejects );
         
         fclose( rejects );
         ftruncate( fileno( temp ), 0 );
         rewind( temp );
         continue;
      }

      rewind( temp );
      do
      {
         if ( fgets( buffer, LN_BUFFER_SIZE, temp ) == NULL )
         {
            fclose( spool );
            fclose( temp );
            fclose( posted );
            strcpy( ln_error_message,
                    "libnews: ln_post_messages: "
                    "premature end of temporary file." );
            return -1;
         }
         fputs( buffer, posted );
      }
      while( strncmp( buffer, ".\r\n", 3 ));

      ++( *count );

      if ( ftruncate( fileno( temp ), 0 ))
      {
         fclose( spool );
         fclose( temp );
         fclose( posted );
         strcpy( ln_error_message,
                 "libnews: ln_post_messages: "
                 "could not truncate temporary file." );
         return -1;
      }

   }

   fclose( spool );
   snprintf( buffer, LN_BUFFER_SIZE, "%s/.peruser_spool/%s-FOLLOW-UPS", home,
             server );
   truncate( buffer, 0 );

   if ( rejects != NULL )
   {
      snprintf( second_buffer, LN_BUFFER_SIZE, "%s/.peruser_spool/.rejects", 
                home );
      rename( second_buffer, buffer );
   }

   fclose( posted );
   fclose( temp );

   return 0;
}

/*
 * Truncates specified spool file to zero length.
 */

int ln_truncate_spool( char *server, char *group, char what )
{
   char *string, buffer[ LN_BUFFER_SIZE ];


   switch( what )
   {
      case 0:
         string = "%s/.peruser_spool/%s-FOLLOW-UPS";
         break;

      case 1:
         string = "%s/.peruser_spool/%s";
         break;

      case 2:
         string = "%s/.peruser_spool/%s-POSTED";
         break;
   }

   snprintf( buffer, LN_BUFFER_SIZE, string, getenv( "HOME" ),
             (( what == 1 ) ? group : server ));
   if ( truncate( buffer, 0 ))
   {
      snprintf( ln_error_message, LN_BUFFER_SIZE,
                "libnews: ln_truncate_spool: could not truncate: %s.",
                buffer );
      return -1;
   }

   if ( what == 1 )
   {
      strcat( buffer, ":read" );
      truncate( buffer, 0 );
   }

   return 0;
}

/*
 * Removes a single article from a spoolfile.
 */

int ln_delete_articles( char *server, char *group, 
                        ln_group_summary *contents, unsigned int total,
                        char what )
{
   char *home, *string,
      buffer[ LN_BUFFER_SIZE ], second_buffer[ LN_BUFFER_SIZE ],
      path[ LN_BUFFER_SIZE ];
   unsigned int i, count;
   FILE *spool, *temp, *read, *temp_read;


   switch( what )
   {
      case 0:
         string = "%s/.peruser_spool/%s-FOLLOW-UPS";
         break;

      case 1:
         string = "%s/.peruser_spool/%s";
         break;

      case 2:
         string = "%s/.peruser_spool/%s-POSTED";
         break;
   }

   snprintf( buffer, LN_BUFFER_SIZE, string, home = getenv( "HOME" ), 
             (( what == 1 ) ? group : server ));
   strcpy( path, buffer );
   
   if (( spool = fopen( buffer, "r" )) == NULL )
   {
      snprintf( ln_error_message, LN_BUFFER_SIZE,
                "libnews: ln_delete_articles: could not open file %s.",
                buffer );
      return -1;
   }

   strcat( buffer, ".tmp" );
   if (( temp = fopen( buffer, "w" )) == NULL )
   {
      snprintf( ln_error_message, LN_BUFFER_SIZE,
                "libnews: ln_delete_articles: could not open file %s.",
                buffer );
      return -1;
   }

   if ( what == 1 )
   {
      strcat( path, ":read" );
      if (( read = fopen( path, "r" )) == NULL )
      {
         snprintf( ln_error_message, LN_BUFFER_SIZE, "libnews: ln_delete_"
                   "articles: could not open file %s.", path );
         return -1;
      }

      strcat( path, ".tmp" );
      if (( temp_read = fopen( path, "w" )) == NULL )
      {
         snprintf( ln_error_message, LN_BUFFER_SIZE, "libnews: ln_delete_"
                   "articles: could not open file %s.", path );
         return -1;
      }
   }
   
   count = 0;
   for( i = 0; i < total; ++ i )
   {
      if ( !( contents[ i ].delete ))
      {
         if ( what == 1 )
            if ( fgets( buffer, LN_BUFFER_SIZE, read ) != NULL )
               fputs( buffer, temp_read );
            else
               fputs( "r\n", temp_read );

         do
         {
            if ( fgets( buffer, LN_BUFFER_SIZE, spool ) == NULL )
            {
               fclose( spool );
               fclose( temp );
               strcpy( ln_error_message, "libnews: ln_delete_articles: "
                     "premature end of spool." );
               return -1;
            }
            fputs( buffer, temp );
         }
         while( strncmp( buffer, ".\r\n", 3 ));
      }
      else
      {
         ++count;

         if ( what == 1 )
            fgets( buffer, LN_BUFFER_SIZE, read );
         
         do
            if ( fgets( buffer, LN_BUFFER_SIZE, spool ) == NULL )
            {
               fclose( spool );
               fclose( temp );
               strcpy( ln_error_message, "libnews: ln_delete_articles: "
                       "premature end of spool." );
               return -1;
            }
         while( strncmp( buffer, ".\r\n", 3 ));
      }
   }

   fclose( spool );
   fclose( temp );

   snprintf( buffer, LN_BUFFER_SIZE, string, home, 
             (( what == 1 ) ? group : server ));
   
   snprintf( second_buffer, LN_BUFFER_SIZE, "%s.tmp", buffer );
   rename( second_buffer, buffer );

   if ( what == 1 )
   {
      fclose( read );
      fclose( temp_read );
      
      strcat( buffer, ":read" );
      rename( path, buffer );
   }
   
   return count;
}
